#include <stdint.h>
#include <string.h>
#include <inttypes.h>
#include <openssl/evp.h>
#include <assert.h>
#include "common.h"
#include "xor.h"
#include "base64.h" // to del
#include <openssl/err.h>
#include "pkcs7.h"

void process_block(uint8_t in[16], uint8_t out[16], const Key_t* const key, bool encrypt)
{
    int ret;

    EVP_CIPHER_CTX ctx;
    EVP_CIPHER_CTX_init(&ctx);
    OP_CHECK(
        EVP_CipherInit_ex(  &ctx, EVP_aes_128_ecb(), NULL, key->key, NULL, encrypt));
    OP_CHECK(  EVP_CIPHER_CTX_iv_length(&ctx) == 0);
    EVP_CIPHER_CTX_set_padding(&ctx, 0);
    OP_CHECK( EVP_CipherUpdate(&ctx, out, &ret, in, 16));
    if(ret != 16) {
        printf("ERROR %d %lu \n", ret, 16);
    }
    OP_CHECK( EVP_CipherFinal_ex(&ctx, out+ret, &ret));
    if(ret != 0) {
        printf("ERROR %d %lu \n", ret, 16);
    }
end:
    EVP_CIPHER_CTX_cleanup(&ctx);
}

Result_t aes_whole_blocks(  CryptoAttribs_t* attribs,
                        const Key_t* const key)
{
    assert(key != NULL);
    assert(attribs != NULL);
    assert(key->len == 16);
    size_t max_size = ROUNDUP_16_BYTES(attribs->input_len) ;
    Result_t res = Result_OK;
    // if output is NULL then allocate same size as input rounded to block size
    if(attribs->output == NULL)
    {
        attribs->output = (uint8_t*) malloc(max_size+1/* +1 because it's needed in final */);
    }

    unsigned blocks = attribs->input_len / 16;
    for(unsigned i=0; i<blocks; i++) {
        process_block(attribs->input+(i*16), attribs->output+attribs->output_len, key, attribs->operation == kEncrypt);
        attribs->output_len += 16;
    }

end:
    return res;
}

Result_t cbc_decrypt(
    CryptoAttribs_t* attribs,
    const Key_t* const key)
{
    if( NULL == attribs->iv )
        return Result_Error;

    if((attribs->input_len % key->len) != 0 )
        return Result_Error;

    if( attribs->iv_len != key->len )
        return Result_Error;

    if( key->len == 0 )
        return Result_Error;

    const size_t bc = (size_t) (attribs->input_len / key->len);

    // Allocate memory for output, if NULL
    if( attribs->output == NULL)
    {
        // same as input + 16 bytes iv
        const size_t len = (ROUNDUP_16_BYTES(attribs->input_len));
        attribs->output = (uint8_t*) malloc(len);
        assert( attribs->output != NULL );
        assert( len != 0 );
    }

    uint8_t* iv = attribs->iv;
    // Allocate struct that will be used for block decryption
    CryptoAttribs_t ecb_attribs;
    CryptoAttribs_t::Init(&ecb_attribs);
    ecb_attribs.output = (uint8_t*) malloc(key->len);
    ecb_attribs.output_len = 0;

    for(int i=0; i<bc; ++i)
    {
        ecb_attribs.input = (attribs->input)+(key->len*i);
        ecb_attribs.input_len = key->len;

        // 1. Decrypt
        ecb_attribs.operation = kDecrypt;
        if( aes_whole_blocks(&ecb_attribs, key) != Result_OK )
            return Result_Error;

        // 2. Xor IV
        unsigned char* out = attribs->output+(key->len)*i;
        xor_repeatedly(
            iv, attribs->iv_len,
            ecb_attribs.output, attribs->iv_len,
            out, key->len);

        // 3. Change iv with block
        iv = ecb_attribs.input;
        attribs->output_len += key->len;
        ecb_attribs.output_len = 0;
    }
    ::free(ecb_attribs.output);

    if(attribs->padding == kPadding_PKCS7)
    {
        uint8_t* unpadded = NULL;
        size_t length;
        pkcs7_unpad(attribs->output, attribs->output_len, &unpadded, &length);
        memcpy(attribs->output, unpadded, length);
        attribs->output_len = length;
        ::free(unpadded);
    }

    return Result_OK;
}

Result_t cbc_encrypt(
    CryptoAttribs_t* attribs,
    const Key_t* const key)
{
    if( attribs->iv_len != key->len )
        return Result_Error;

    if( key->len == 0 )
        return Result_Error;

    uint8_t* local_input = NULL;
    size_t local_input_len = 0;
    if( attribs->padding == kPadding_PKCS7 )
    {
        // TODO: it should round up to key length not necsairlly 16
        local_input_len = ROUNDUP_16_BYTES(attribs->input_len);
        // one block extra if size aligned
        if(local_input_len == attribs->input_len)
            local_input_len += 16;
        pkcs7_pad(attribs->input, attribs->input_len, &local_input, local_input_len);
    }
    else
    {
        if((attribs->input_len % key->len) != 0 )
            return Result_Error;
        local_input_len = attribs->input_len;
        local_input = (uint8_t*) malloc(local_input_len);
        memcpy(local_input, attribs->input, local_input_len);
    }

    const size_t bc = (size_t)( local_input_len / key->len );
    const size_t bs = attribs->iv_len;
    assert(bs == 16);

    if( attribs->output == NULL)
    {
        // same as input + 16 bytes iv
        attribs->output =  (uint8_t*) malloc(
                    (ROUNDUP_16_BYTES(local_input_len)+16) );
        assert( attribs->output != NULL );
    }
    uint8_t* iv = attribs->iv;
    // Initialize attribs for ECB block encryption
    CryptoAttribs_t ecb_attribs;
    CryptoAttribs_t::Init(&ecb_attribs);
    ecb_attribs.input = (uint8_t*)malloc(bs);
    ecb_attribs.input_len = 0;
    ecb_attribs.output = (uint8_t*)malloc(bs);
    ecb_attribs.output_len = 0;
    Result_t ret = Result_OK;
    for(int i=0; i<bc; ++i)
    {
        uint8_t* buffptr = (local_input)+(bs*i);

        // 1. Xor IV
        xor_repeatedly(
            iv, bs,
            buffptr, bs,
            ecb_attribs.input, bs);
        ecb_attribs.input_len = bs;

        // 2. Encrypt
        ecb_attribs.operation = kEncrypt;
        if( aes_whole_blocks(&ecb_attribs, key) != Result_OK )
        {
            ret = Result_Error;
            break;
        }
        assert(ecb_attribs.output != NULL);
        // Encrypted block is my new IV
        iv = ecb_attribs.output;

        // 3. Change iv with block
        memcpy((attribs->output)+(bs*i),
                ecb_attribs.output,
                ecb_attribs.output_len);
        assert(ecb_attribs.output_len == 16);
        attribs->output_len += ecb_attribs.output_len;
        ecb_attribs.output_len = 0;
    }
    CryptoAttribs_t::Free(&ecb_attribs);
    ::free(local_input);
    return ret;
}

Result_t ecb_encrypt( CryptoAttribs_t* attribs, const Key_t* const key )
{
    attribs->operation = kEncrypt;
    Result_t res = Result_OK;

    if( kPadding_PKCS7==attribs->padding )
    {
        CryptoAttribs_t padded_attribs;
        CryptoAttribs_t::Init(&padded_attribs);
        do
        {
            // TODO: it should round up to key length not necsairlly 16
            padded_attribs.input_len = ROUNDUP_16_BYTES(attribs->input_len);

            // one block extra if size aligned
            if(padded_attribs.input_len == attribs->input_len)
                padded_attribs.input_len += 16;
            if( !pkcs7_pad(attribs->input, attribs->input_len, &padded_attribs.input, padded_attribs.input_len) )
            {
                res = Result_Error;
                break;
            }

            padded_attribs.padding = kPadding_None;
            res = aes_whole_blocks(&padded_attribs, key);
            if(NULL==attribs->output)
            {
                attribs->output = (uint8_t*)malloc(padded_attribs.output_len);
            }

            memcpy(attribs->output, padded_attribs.output, padded_attribs.output_len);
            attribs->output_len = padded_attribs.output_len;
        } while(false);
        CryptoAttribs_t::Free(&padded_attribs);
    }
    else
    {
        res = aes_whole_blocks(attribs, key);
    }
    return res;
}

Result_t ecb_decrypt( CryptoAttribs_t* attribs, const Key_t* const key )
{
    Result_t res = Result_OK;
    attribs->operation = kDecrypt;

    res = aes_whole_blocks(attribs, key);
    if( kPadding_PKCS7==attribs->padding )
    {
        size_t output_len=0;
        uint8_t* output=NULL;
        if(!pkcs7_unpad(attribs->output, attribs->output_len, &output, &output_len))
        {
            return Result_Error;
        }
        memcpy(attribs->output, output, output_len);
        attribs->output_len=output_len;
    }
    return res;
}
