#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>


/* -----------------------------------------------------------------------------
 * @brief   Produces PKCS7 padded buffer from unpadded buffer
 *
 * @param   i_buff: unpadded buffer
 *          i_len : length of unpadded buffer
 *          o_buff: padded output. If pointer *o_buff NULL then memory is allocated
 *          i_obuff_len: requested size of the output buffer
 *
 * @returns false if:
 *              * pad value is bigger then 0xFF. In such case o_buff is unchanged.
 *              * i_obuff_len < i_len
 *              * o_buff: NULL
 *          Otherwise true.
 *
-------------------------------------------------------------------------------- */
bool pkcs7_pad(const uint8_t* i_buff, size_t i_len, uint8_t** o_buff, size_t i_obuff_len )
{
    if( (i_obuff_len < i_len ) ||
        (i_obuff_len - i_len > 0xFF) ||
        (o_buff == NULL) )
    {
        return false;
    }
    if(NULL == *o_buff )
    {
        *o_buff = (uint8_t*) malloc(i_obuff_len);
    }
    const unsigned pad_value = i_obuff_len - i_len;
    memset(*o_buff, pad_value, i_obuff_len);
    memcpy(*o_buff, i_buff, i_len);

    return true;
}

/* -----------------------------------------------------------------------------
 * @brief   Removes padding from PKCS7 padded buffer
 *
 * @param   i_padded_buf: padded buffer
 *          i_len : length of padded buffer
 *          o_unpad_buf: unpadded output. If pointer *o_buff NULL then memory is allocated
 *          unpad_len: length of unpadded buffer
 *
 * @returns false if:
 *              * any of pointers NULL
 *              * i_len == 0
 *              * pad_value > 0xFF
 *          Otherwise true.
 *
-------------------------------------------------------------------------------- */
bool pkcs7_unpad(   const uint8_t* i_padded_buf,
                    size_t i_len,
                    uint8_t** o_unpad_buf,
                    size_t* unpad_len)
{
    if( i_padded_buf== NULL ||
        unpad_len   == NULL ||
        o_unpad_buf == NULL ||
        i_len       == 0)
    {
        return false;
    }

    const unsigned pad_val = i_padded_buf[i_len - 1];
    if(pad_val > 0xFF)
    {
        return false;
    }

    *unpad_len = i_len - pad_val;
    if(NULL==*o_unpad_buf)
    {
        *o_unpad_buf = (uint8_t*) malloc(*unpad_len);
    }

    memcpy(*o_unpad_buf, i_padded_buf, *unpad_len);
    return true;
}